// Author: Sol Boucher <sboucher@cmu.edu>

#include <netinet/in.h>
#include <sys/socket.h>
#include <ctype.h>
#include <netdb.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#define HOSTNAME_LEN 64
#define PORTNUM_LEN 8
#define QUEUE_LEN 8

int main(int argc, char **argv) {
	if(argc != 2) {
		printf("USAGE: %s <port>\n", argv[0]);
		return 1;
	}
	int port = atoi(argv[1]);

	// 1. socket(family, socktype, transport)
	int sockfd;
	if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket()");
		return 2;
	}

	// 2. bind(fd, addr, size)
	struct sockaddr_in addr = {
		.sin_family = AF_INET,
		.sin_port = htons(port),
		.sin_addr = {INADDR_ANY},
	};
	if(bind(sockfd, (struct sockaddr *) &addr, sizeof addr)) {
		perror("bind()");
		return 3;
	}

	// 3. listen(fd, len)
	if(listen(sockfd, QUEUE_LEN)) {
		perror("listen()");
		return 4;
	}

	while(true) {
		// 4. accept(fd, addr, size)
		struct sockaddr_in peer;
		socklen_t plen = sizeof peer;
		int connfd;
		if((connfd = accept(sockfd, (struct sockaddr *) &peer, &plen)) < 0) {
			perror("accept()");
			continue;
		}

		// 5. getnameinfo(addr, size, name, size, port, size, flags)
		char hostname[HOSTNAME_LEN];
		char portnum[PORTNUM_LEN];
		int stat;
		if((stat = getnameinfo((struct sockaddr *) &peer, plen, hostname, sizeof hostname, portnum, sizeof portnum, 0))) {
			fprintf(stderr, "%s: %s\n", "getnameinfo()", gai_strerror(stat));
			continue;
		}

		printf("# incoming connection from %s:%s\n", hostname, portnum);

		FILE *conn = fdopen(connfd, "r");
		if(!conn) {
			perror("fdopen()");
			continue;
		}

		int chr;
		while((chr = getc(conn)) != EOF)
			if(isprint(chr))
				putchar(chr);
			else
				printf("\\x%x", chr);
		putchar('\n');

		fclose(conn);

		puts("# connection closed by remote host");
	}

	return 0;
}
